# -*- coding: utf-8 -*-

from __future__ import division

import collections
import warnings

import numpy as np

__all__ = 'qn', 'Cumulant', 'Sampler'


def qn(phi, *n):
    """
    Calculate the complex flow vector `Q_n`.

    :param array-like phi: Azimuthal angles.

    :param int n: One or more harmonics to calculate.

    :returns:
        A single complex number if only one ``n`` was given or a complex array
        for multiple ``n``.

    """
    phi = np.ravel(phi)
    n = np.asarray(n)

    i_n_phi = np.zeros((n.size, phi.size), dtype=complex)
    np.outer(n, phi, out=i_n_phi.imag)

    qn = np.exp(i_n_phi, out=i_n_phi).sum(axis=1)
    if qn.size == 1:
        qn = qn[0]

    return qn


class Cumulant(object):
    r"""
    Multi-particle flow correlations and cumulants for an ensemble of events.

    Each argument must be an array-like object where each element corresponds
    to an event: ``multiplicities`` is an array containing event-by-event
    multiplicities, ``q2`` is an array of the same size containing the `Q_2`
    vectors for the same set of events, and so on.

    :param array-like multiplicities: Event-by-event multiplicities.

    :param array-like q2, q3, ...:
        `Q_n` vectors as positional arguments.

    :param array-like qn_kwargs:
        `Q_n` vectors as keyword arguments.

    Member functions ``correlation``, ``cumulant``, and ``flow`` compute the
    correlation function, cumulant, and flow coefficient, respectively.  Each
    takes two arguments ``(n, k)`` where ``n`` (positive integer) is the
    anisotropy order and ``k`` (even positive integer) is the correlation
    order.  A given `(n, k)` requires flow vectors `Q_n, Q_{2n}, \ldots,
    Q_{nk/2}`, e.g. ``(2, 4)`` requires ``q2, q4``.  Functions will raise
    ``ValueError`` if they don't have the required flow vectors.  Currently
    ``k=2`` and ``k=4`` are implemented; ``k=6`` is planned.

    """
    def __init__(self, multiplicities, *qn, **qn_kwargs):
        # Multiplicity must be stored as floating point because the large
        # powers of M calculated in n-particle correlations can overflow
        # integers, e.g. 2000^6 > 2^64.
        self._M = np.asarray(multiplicities, dtype=float).ravel()

        try:
            qn_dict = {int(k.lstrip('q')): v for k, v in qn_kwargs.items()}
        except ValueError:
            raise TypeError("Keyword parameters must have the form 'qN' "
                            "where N is an integer.")
        qn_dict.update(enumerate(qn, start=2))
        self._qn = {k: np.asarray(v, dtype=complex).ravel()
                    for k, v in qn_dict.items()
                    if v is not None}

        self._corr = collections.defaultdict(dict)
        self._corr_err = collections.defaultdict(dict)

    def _get_qn(self, n):
        try:
            return self._qn[n]
        except KeyError:
            raise ValueError(
                'q_{} is required for this calculation but was not provided.'
                .format(n)
            )

    def _calculate_corr(self, n, k, error=False):
        # skip if this correlation was already calculated
        if k in self._corr[n]:
            return

        # TODO: put these functions in a _util module
        M = self._M
        Msum = np.einsum('i->', M)  # fast sum(M)
        Msqsum = np.inner(M, M)  # fast sum(M^2)

        qn = self._get_qn(n)
        qnsqsum = np.vdot(qn, qn).real  # fast sum(|qn|^2)

        if k == 2:
            self._corr[n][k] = (qnsqsum - Msum) / (Msqsum - Msum)

        elif k == 4:
            q2n = self._get_qn(2*n)
            q2nsqsum = np.vdot(q2n, q2n).real
            qnsq = np.square(qn.real) + np.square(qn.imag)
            qnto4sum = np.inner(qnsq, qnsq)
            self._corr[n][k] = (
                qnto4sum +
                q2nsqsum -
                2*np.inner(q2n, np.square(qn.conj())).real -
                2*np.sum(2*(M-2)*qnsq - M*(M-3))
            ) / np.sum(M*(M-1)*(M-2)*(M-3))

        # TODO: k == 6

        else:
            raise ValueError('Unknown k: {}.'.format(k))

    def _calculate_corr_err(self, n, k):
        # skip if this error was already calculated
        if k in self._corr_err[n]:
            return

        if k == 2:
            M = self._M
            weights = M*(M-1)
            w1 = np.einsum('i->', weights)  # sum of weights
            w2 = np.inner(weights, weights)  # sum of squared weights

            # event-by-event differences from mean: <2>_i - <<2>>
            qn = self._get_qn(n)
            diff = (np.real(qn*qn.conj())-M)/weights - self._corr[n][k]

            # weighted variance
            var = np.sum(weights*np.square(diff)) / w1

            # unbiased variance s^2
            ssq = var / (1. - w2/(w1*w1))

            # final error of <<2>>
            self._corr_err[n][k] = np.sqrt(w2*ssq)/w1

        else:
            raise ValueError('Error is only implemented for k == 2.')

    def correlation(self, n, k, error=False):
        r"""
        Calculate `\langle k \rangle_n`,
        the `k`-particle correlation function for `n`\ th-order anisotropy.

        :param int n: Anisotropy order.
        :param int k: Correlation order.

        :param bool error:
            Whether to calculate statistical error
            (for `\langle 2 \rangle` only).
            If true, return a tuple ``(corr, corr_error)``.

        """
        self._calculate_corr(n, k)
        corr_nk = self._corr[n][k]

        if error:
            self._calculate_corr_err(n, k)
            return corr_nk, self._corr_err[n][k]
        else:
            return corr_nk

    def cumulant(self, n, k, error=False):
        r"""
        Calculate `c_n\{k\}`,
        the `k`-particle cumulant for `n`\ th-order anisotropy.

        :param int n: Anisotropy order.
        :param int k: Correlation order.

        :param bool error:
            Whether to calculate statistical error (for `c_n\{2\}` only).
            If true, return a tuple ``(cn2, cn2_error)``.

        """
        corr_nk = self.correlation(n, k, error=error)

        if k == 2:
            return corr_nk
        elif k == 4:
            corr_n2 = self.correlation(n, 2)
            return corr_nk - 2*corr_n2*corr_n2

    _cnk_prefactor = {2: 1, 4: -1}

    def flow(self, n, k, error=False, imaginary='nan'):
        r"""
        Calculate `v_n\{k\}`,
        the estimate of flow coefficient `v_n` from the `k`-particle cumulant.

        :param int n: Anisotropy order.
        :param int k: Correlation order.

        :param bool error:
            Whether to calculate statistical error (for `v_n\{2\}` only).
            If true, return a tuple ``(vn2, vn2_error)``.

        :param str imaginary: (optional)
            Determines behavior when the computed flow is imaginary:

            - ``'nan'`` (default) -- Return NaN and raise a ``RuntimeWarning``.
            - ``'negative'`` -- Return the negative absolute value.
            - ``'zero'`` -- Return ``0.0``.

        """
        cnk = self.cumulant(n, k, error=error)

        if error:
            cnk, cnk_err = cnk

        vnk_to_k = self._cnk_prefactor[k] * cnk
        kinv = 1/k

        if vnk_to_k >= 0:
            vnk = vnk_to_k**kinv
        else:
            if imaginary == 'negative':
                vnk = -1*(-vnk_to_k)**kinv
            elif imaginary == 'zero':
                vnk = 0.
            else:
                warnings.warn('Imaginary flow: returning NaN.', RuntimeWarning)
                vnk = float('nan')

        if k == 2 and error:
            return vnk, .5/np.sqrt(abs(cnk)) * cnk_err
        else:
            return vnk


class Sampler(object):
    r"""
    Random flow event generator.

    A ``Sampler`` object represents an event with specified flow coefficients
    `v_n`.  It computes and randomly samples `dN/d\phi` as a probability
    density function (PDF).

    :param float v2, v3, ...:
        Flow coefficients as positional arguments.

    :param float vn_kwargs:
        Flow coefficients as keyword arguments.

    If no ``vn`` arguments are given, ``Sampler`` will have zero flow, i.e.
    ``Sampler.pdf()`` will be flat and ``Sampler.sample()`` will generate
    uniform numbers from `[-\pi, \pi)`.

    """
    def __init__(self, *vn, **vn_kwargs):
        if not vn and not vn_kwargs:
            self._n = self._vn = None
        else:
            try:
                vn_dict = {int(k.lstrip('v')): v for k, v in vn_kwargs.items()}
            except ValueError:
                raise TypeError("Keyword parameters must have the form 'vN' "
                                "where N is an integer.")
            vn_dict.update((k, v) for k, v in enumerate(vn, start=2)
                           if v is not None and v != 0.)
            kwargs = dict(dtype=float, count=len(vn_dict))
            self._n = np.fromiter(vn_dict.keys(), **kwargs)
            self._vn = np.fromiter(vn_dict.values(), **kwargs)

    def _pdf(self, phi):
        """
        Evaluate the _unnormalized_ flow PDF.

        """
        pdf = np.inner(self._vn, np.cos(np.outer(phi, self._n)))
        pdf *= 2.
        pdf += 1.

        return pdf

    @staticmethod
    def _uniform_phi(M):
        """
        Generate M random numbers in [-pi, pi).

        """
        return np.random.uniform(-np.pi, np.pi, M)

    def pdf(self, phi):
        r"""
        Evaluate the flow PDF `dN/d\phi`.

        :param array-like phi: Azimuthal angles.

        :returns: The flow PDF evaluated at ``phi``.

        """
        if self._n is None:
            pdf = np.empty_like(phi)
            pdf.fill(.5/np.pi)
            return pdf

        phi = np.asarray(phi)

        pdf = self._pdf(phi)
        pdf /= 2.*np.pi

        return pdf

    def sample(self, multiplicity):
        r"""
        Randomly sample azimuthal angles `\phi`.

        :param int multiplicity: Number to sample.

        :returns: Array of sampled angles.

        """
        if self._n is None:
            return self._uniform_phi(multiplicity)

        # Since the flow PDF does not have an analytic inverse CDF, I use a
        # simple accept-reject sampling algorithm.  This is reasonably
        # efficient since for normal-sized vn, the PDF is close to flat.  Now
        # due to the overhead of Python functions, it's desirable to minimize
        # the number of calls to the random number generator.  Therefore I
        # sample numbers in chunks; most of the time only one or two chunks
        # should be needed.  Eventually, I might rewrite this with Cython, but
        # it's fast enough for now.

        N = 0  # number of phi that have been sampled
        phi = np.empty(multiplicity)  # allocate array for phi
        pdf_max = 1 + 2*self._vn.sum()  # sampling efficiency ~ 1/pdf_max

        while N < multiplicity:
            n_remaining = multiplicity - N
            n_to_sample = int(1.03*pdf_max*n_remaining)
            phi_chunk = self._uniform_phi(n_to_sample)
            phi_chunk = phi_chunk[self._pdf(phi_chunk) >
                                  np.random.uniform(0, pdf_max, n_to_sample)]
            K = min(phi_chunk.size, n_remaining)  # number of phi to take
            phi[N:N+K] = phi_chunk[:K]
            N += K

        return phi



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal

from .. import initial


def test_ic():
    """initial conditions"""

    ic = initial.IC((
        (0, 0, 0, 0),
        (0, 1, 1, 0),
        (0, 1, 1, 0),
        (0, 0, 0, 0),
    ), 1)

    assert_almost_equal(ic.sum(), 4, err_msg='Incorrect total profile.')

    assert_array_almost_equal(ic.cm(), (0, 0),
                              err_msg='Incorrect center of mass.')

    for n in range(1, 10):
        ecc_correct = 0 if n % 4 else 1
        assert_almost_equal(ic.ecc(n), ecc_correct,
                            err_msg='Incorrect epsilon_{}.'.format(n))

    ic = initial.IC((
        (0, 0, 0, 0, 0),
        (0, 2, 0, 1, 0),
        (0, 0, 0, 0, 0),
        (0, 3, 0, 4, 0),
        (0, 0, 0, 0, 0),
    ), .1)

    assert_almost_equal(ic.sum(), .1, err_msg='Incorrect total profile.')

    assert_array_almost_equal(ic.cm(), (0, -.04),
                              err_msg='Incorrect center of mass.')

    S = np.array((1, 2, 3, 4), dtype=float)
    X = np.array((.1, -.1, -.1, .1))
    Y = np.array((.1, .1, -.1, -.1))
    Y += .04
    Rsq = X*X + Y*Y

    for n in range(1, 10):
        ecc = abs(np.average(
            np.exp(1j*n*np.arctan2(Y, X)),
            weights=S*Rsq**(.5*n)
        ))
        assert_almost_equal(ic.ecc(n), ecc,
                            err_msg='Incorrect epsilon_{}.'.format(n))

    # TODO: test non-square grids and grid spacing



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

import hic

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='hic',
    version=hic.__version__,
    description='Tools for analyzing heavy-ion collision simulations.',
    long_description=long_description,
    author='Jonah Bernhard',
    author_email='jonah.bernhard@gmail.com',
    url='https://github.com/Duke-QCD/hic',
    license='MIT',
    packages=['hic', 'hic.test'],
    install_requires=['numpy'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Physics'
    ]
)



# -*- coding: utf-8 -*-

from __future__ import division

import itertools
import warnings

import numpy as np
from nose.tools import assert_raises

from .. import flow


def test_qn():
    """flow vectors q_n"""

    # This is _almost_ unnecessary.  It would be tough to screw up q_n.

    # q_n(0) = 1
    q = flow.qn(0, 2)
    assert q == 1., \
        'Incorrect single-particle q_n.\n{} != 1'.format(q)

    # q_3(isotropic phi) = -1
    q = flow.qn((0, np.pi/3, -np.pi/3), 3)
    assert q == -1., \
        'Incorrect isotropic q_3.\n{} != -1'.format(q)

    n_ = 2, 3
    phi = np.random.uniform(-np.pi, np.pi, 10)
    assert np.allclose(
        flow.qn(phi, *n_),
        [flow.qn(phi, n) for n in n_]
    ), 'Simultaneous qn do not agree with individuals.'


def assert_close(a, b, msg='', tol=1e-15):
    assert abs(a - b) < tol, \
        '{}\n{} != {}'.format(msg, a, b)


def test_cumulant():
    """flow correlations and cumulants"""

    # Compare n-particle correlations to explicitly calculating all
    # n-permutations.
    M = 8
    phi = 2*np.pi*np.random.rand(8)

    corr2_true = np.exp(2j*np.array([
        (p1-p2) for p1, p2 in itertools.permutations(phi, 2)
    ])).mean()
    corr4_true = np.exp(2j*np.array([
        (p1+p2-p3-p4) for p1, p2, p3, p4 in itertools.permutations(phi, 4)
    ])).mean()

    q2, q4 = (np.exp(1j*n*phi).sum() for n in (2, 4))
    vnk = flow.Cumulant(M, q2, None, q4)
    assert sorted(vnk._qn.keys()) == [2, 4], \
        'Incorrect parsing of qn arguments.'
    corr2, corr4 = (vnk.correlation(2, n) for n in (2, 4))

    assert_close(corr2, corr2_true, 'Incorrect 2-particle correlation.')
    assert_close(corr4, corr4_true, 'Incorrect 4-particle correlation.')

    # Now verify cumulants.
    c2_true = corr2_true
    c4_true = corr4_true - 2*corr2_true*corr2_true

    c2, c4 = (vnk.cumulant(2, n) for n in (2, 4))

    assert_close(c2, c2_true, 'Incorrect 2-particle cumulant.')
    assert_close(c4, c4_true, 'Incorrect 4-particle cumulant.')

    # Fabricate q_n so that v2 is positive, v3 vanishes, and v4 is imaginary.
    Nev = 10
    M = 100
    qfluct = M**.5*np.ones(Nev)
    qn = 1.1*qfluct, qfluct, 0.9*qfluct
    vnk = flow.Cumulant(M*np.ones(Nev), *qn)

    v22 = vnk.flow(2, 2)
    assert v22 > 0, \
        'This v_2{2} must be positive.\n{}'.format(v22)

    v32 = vnk.flow(3, 2)
    assert v32 == 0, \
        'This v_3{2} must vanish.\n{}'.format(v32)

    with warnings.catch_warnings(record=True) as w:
        v42_nan = vnk.flow(4, 2)
        assert len(w) == 1 and issubclass(w[0].category, RuntimeWarning), \
            'There should be RuntimeWarning here.'
    assert np.isnan(v42_nan), \
        'This v_4{2} must be NaN.\n{}'.format(v42_nan)

    v42_zero = vnk.flow(4, 2, imaginary='zero')
    assert v42_zero == 0, \
        'This v_4{2} must be zero.\n{}'.format(v42_zero)

    v42_negative = vnk.flow(4, 2, imaginary='negative')
    assert v42_negative < 0, \
        'This v_4{2} must be negative.\n{}'.format(v42_negative)

    # Test statistical error.
    v22, v22err = vnk.flow(2, 2, error=True)
    assert v22err < 1e-16, \
        'This v_2{2} error must vanish.\n{}'.format(v22err)
    assert (v22, v22err) == vnk.flow(2, 2, error=True), \
        'Cached result does not match original.'

    Nev = 5
    M = 10
    w1 = Nev*M*(M-1)
    w2 = Nev*np.square(M*(M-1))

    phi_events = 2*np.pi*np.random.rand(Nev, M)
    q2, q4 = (np.array([flow.qn(phi, n) for phi in phi_events])
              for n in (2, 4))
    vnk = flow.Cumulant(np.full(Nev, M), q2=q2, q4=q4)

    for n in 2, 4:
        corr_ebe = np.array([
            np.exp(n*1j*np.array([
                (p1-p2) for p1, p2 in itertools.permutations(phi, 2)
            ])).mean()
            for phi in phi_events])

        corr_true = corr_ebe.mean()
        corr_err_true = np.sqrt(w2/(w1*w1 - w2)) * corr_ebe.std()

        corr, corr_err = vnk.correlation(n, 2, error=True)
        assert_close(corr, corr_true,
                     'Incorrect {}-particle correlation.'.format(n))
        assert_close(corr_err, corr_err_true,
                     'Incorrect {}-particle correlation error.'.format(n))

    # Test bad arguments.

    # invalid k
    assert_raises(ValueError, vnk.correlation, 2, 3)

    # error only implemented for k == 2
    assert_raises(ValueError, vnk.correlation, 2, 4, error=True)

    # missing qn
    assert_raises(ValueError, vnk.cumulant, 4, 4)

    # bad class keyword args
    assert_raises(TypeError, flow.Cumulant, M, v2=q2)


def test_sampler():
    """flow random sampling"""

    # zero flow
    sampler = flow.Sampler()
    assert sampler._n is None and sampler._vn is None, \
        'Incorrect vn argument parsing.'

    phi = np.random.rand(5)
    assert np.allclose(sampler.pdf(phi), 1/(2*np.pi)), \
        'Incorrect uniform flow pdf.'

    M = 1000
    phi = sampler.sample(M)
    vnobs = np.cos(np.outer(np.arange(2, 6), phi)).mean(axis=1)
    assert phi.size == M, \
        'Incorrect number of particles.'
    assert np.all((phi >= -np.pi) & (phi < np.pi)), \
        'Sampled angles are not in [-pi, pi).'
    assert np.all(np.abs(vnobs) < 3.*M**(-.5)), \
        'Flows are not within statistical fluctuation.'

    # nonzero flow
    vn = v2, v3, v4, v5 = .1, .05, 0., .02
    for sampler in (
            flow.Sampler(v2, v3, None, v5),
            flow.Sampler(v2, v3, v5=v5),
            flow.Sampler(v5=v5, v2=v2, v3=v3)
    ):
        assert (
            np.all(sampler._n == (2, 3, 5)) and
            np.all(sampler._vn == (v2, v3, v5))
        ), 'Incorrect vn argument parsing.'

    pdf = (1 +
           2*v2*np.cos(2*phi) +
           2*v3*np.cos(3*phi) +
           2*v5*np.cos(5*phi)
           )/(2*np.pi)
    assert np.allclose(sampler.pdf(phi), pdf), \
        'Incorrect nonuniform flow pdf.'

    M = 1000
    phi = sampler.sample(M)
    vnobs = np.cos(np.outer(np.arange(2, 6), phi)).mean(axis=1)
    assert phi.size == M, \
        'Incorrect number of particles.'
    assert np.all((phi >= -np.pi) & (phi < np.pi)), \
        'Sampled angles are not in [-pi, pi).'
    assert np.all(np.abs(vn - vnobs) < 3.*M**(-.5)), \
        'Flows are not within statistical fluctuation.'

    # bad args
    assert_raises(TypeError, flow.Sampler, M, x2=.1)



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

__all__ = 'IC',


class IC(object):
    """
    Initial condition entropy or energy density profile.

    :param array-like profile:
        The IC profile as a block-style grid.

    :param dxy:
        Size of each grid cell in fm, either a single value ``dxy = dx = dy``
        or a pair ``dxy = (dx, dy)``.
    :type dxy: float or pair of floats

    """
    def __init__(self, profile, dxy):
        self._profile = np.asarray(profile, dtype=float)

        # save (x, y) steps
        try:
            self._dx, self._dy = dxy
        except (TypeError, ValueError):
            self._dx = self._dy = dxy

        # save (x, y) max
        ny, nx = self._profile.shape
        xmax = .5*self._dx*(nx - 1.)
        ymax = .5*self._dy*(ny - 1.)
        self._xymax = xmax, ymax

        # calculate and save center of mass
        X = np.linspace(-xmax, xmax, nx)
        Y = np.linspace(ymax, -ymax, ny)
        cm = np.array((
            np.inner(X, self._profile.sum(axis=0)),
            np.inner(Y, self._profile.sum(axis=1))
        ))
        cm /= self._profile.sum()
        self._cm = cm

    def sum(self):
        """
        Total entropy or energy.

        """
        return self._profile.sum() * self._dx * self._dy

    def cm(self):
        """
        Center of mass coordinates, assuming the middle of the profile is
        (0, 0).

        """
        return self._cm

    def ecc(self, n):
        r"""
        Calculate eccentricity harmonic `\varepsilon_n`.

        :param int n: Eccentricity order.

        """
        ny, nx = self._profile.shape
        xmax, ymax = self._xymax
        xcm, ycm = self._cm

        # create (X, Y) grids relative to CM
        Y, X = np.mgrid[ymax:-ymax:1j*ny, -xmax:xmax:1j*nx]
        X -= xcm
        Y -= ycm

        # create grid of weights = profile * R^n
        Rsq = X*X + Y*Y
        if n == 1:
            W = np.sqrt(Rsq, out=Rsq)
        elif n == 2:
            W = Rsq
        else:
            if n & 1:  # odd n
                W = np.sqrt(Rsq)
            else:  # even n
                W = np.copy(Rsq)
            # multiply by R^2 until W = R^n
            for _ in range(int((n-1)/2)):
                W *= Rsq
        W *= self._profile

        # create grid of e^{i*n*phi} * W
        i_n_phi = np.zeros_like(X, dtype=complex)
        np.arctan2(Y, X, out=i_n_phi.imag)
        i_n_phi.imag *= n
        exp_phi = np.exp(i_n_phi, out=i_n_phi)
        exp_phi *= W

        return abs(exp_phi.sum()) / W.sum()
